/*
 * Copyright (C) 2003 John David Ratliff
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;

import java.net.Socket;
import java.net.ServerSocket;

/**
 * Server - a simple tcp server which waits on a signel from a client
 *          (anything send by the client will do), and sends a pre-loaded
 *          file back to the client.
 */
public class Server {
    private ServerSocket socket;
    private boolean running = true, single;
    private byte[] dataFile;
    
    /**
     * The default server port to use.
     */
    public final static int DEFAULT_PORT = 64201;
    
    /**
     * The default file to send to clients.
     */
    public final static String DEFAULT_FILE = "tcp_spec.txt";

    /**
     * Creates and initializes a Server object.
     *
     * @param port The server port to listen on.
     * @param file The file to send to clients.
     * @param single Use true to send one character at a time, false otherwise.
     */    
    public Server(int port, String file, boolean single) {
        try {
            // set write mode
            this.single = single;
            
            // open the server socket
            this.socket = new ServerSocket(port);
            
            // read the file into the byte array first
            if (!this.readDataFile(file)) {
                this.running = false;
            }
            
            // wait for client connections
            this.waitForClients();
        } catch (IOException e) {
            // if we can't open the Server Socket, we can't run the server
            this.running = false;
            System.out.println("Unable to start the server.");
        }
    }

    /**
     * Default server initializer. Creates a server using the default port and file.
     */    
    public Server() {
        this(DEFAULT_PORT, DEFAULT_FILE, false);
    }

    /**
     * Reads the data file into a byte array for quick supply to clients.
     *
     * @param filename The file to read.
     * @return true on success, false otherwise.
     */    
    private boolean readDataFile(String filename) {
        File file = new File(filename);
        FileInputStream reader;
        int length = (int)file.length();
        
        try {
            // create the file input stream
            reader = new FileInputStream(file);
            
            // create the byte array
            this.dataFile = new byte[length];
            
            // read the file contents
            reader.read(dataFile, 0, dataFile.length);
            
            // close the file stream
            reader.close();
        } catch (FileNotFoundException e) {
            System.out.println("File not found.");
            return false;
        } catch (IOException e) {
            System.out.println("Unable to read from file...");
            e.printStackTrace();
            return false;
        }
        
        return true;
    }

    /**
     * Waits for client connections on the socket. Spawns a thread to handle
     * each incoming client connection.
     */    
    private void waitForClients() {
        Socket client;
        Thread thread;
        Worker worker;

        // If initialization failed, don't go into the accept loop
        if (!running) {
            System.out.println("Unable to start the server.");
            return;
        }

        // the main client accept loop
        for (;;) {
            try {
                // accept the client connection
                client = this.socket.accept();
		
		// create a worker to handle the connection
                worker = new Worker(client, dataFile, single);

		// spawn a thread for the worker
                thread = new Thread(worker);
                thread.start();
            } catch (IOException e) {
                System.out.println("Unable to complete client connection...");
                e.printStackTrace();
            }
        }
    }

    /**
     * Interface to the Server program.
     *
     * @param args Supply the port and filename, or leave blank to use the defaults.
     *             Supply a third option to use the single char write mode.
     */    
    public static void main(String[] args) {
        Server server;
        
        if (args.length == 0) {
            server = new Server();
        } else if (args.length == 2) {
            server = new Server(Integer.parseInt(args[0]), args[1], false);
        } else if (args.length == 3) {
            System.out.println("Setting single character write mode.");
            server = new Server(Integer.parseInt(args[0]), args[1], true);
        } else {
            System.out.println("syntax: java Server port file_name [ single ]");
            System.out.println("ex: java Server 64201 tcp_spec.txt\n\tNormal write mode.");
            System.out.println("ex: java Server 64201 tcp_spec.txt single\n\tSingle char write mode.");
        }
    }
}
